from typing import Callable

import pytdbot

from ..filters import Filter


class Handler:
    r"""A handler class."""

    def __init__(
        self,
        func: Callable,
        update_type: str,
        filter: Filter = None,
        position: int = None,
        inner_object: bool = False,
        timeout: float = None,
        is_from_plugin: bool = False,
    ) -> None:
        self.func = func
        self.update_type = update_type
        self.filter = filter
        self.position = position
        self.inner_object = inner_object
        self.timeout = timeout
        self.is_from_plugin = is_from_plugin

    def __call__(self, client: "pytdbot.Client", update: "pytdbot.types.Update"):
        return self.func(client, update)

    def __str__(self) -> str:
        return f"Handler(func={self.func}, update_type={self.update_type}, filter={self.filter}, position={self.position}, inner_object={self.inner_object}, timeout={self.timeout}, is_from_plugin={self.is_from_plugin})"

    def __repr__(self) -> str:
        return str(self)



import logging
from asyncio import iscoroutinefunction
from typing import Callable

import pytdbot

from .handler import Handler
from .td_updates import Updates

logger = logging.getLogger(__name__)


class Decorators(Updates):
    """Decorators class."""

    def initializer(
        self: "pytdbot.Client" = None,
        filters: "pytdbot.filters.Filter" = None,
        position: int = None,
        inner_object: bool = False,
        timeout: float = None,
    ) -> None:
        r"""A decorator to initialize an event object before running other handlers

        Parameters:
            filters (:class:`~pytdbot.filters.Filter`, *optional*):
                An update filter

            position (``int``, *optional*):
                The function position in initializers list. Default is ``None`` (append)

            inner_object (``bool``, *optional*):
                Wether to pass an inner object of update or not; for example ``UpdateNewMessage.message``. Default is ``False``

            timeout (``float``, *optional*):
                Max execution time for the initializer before it timeout. Default is ``None``

        Raises:
            :py:class:`TypeError`
        """

        def decorator(func: Callable) -> Callable:
            if hasattr(func, "_handler"):
                return func
            elif isinstance(self, pytdbot.Client):
                if iscoroutinefunction(func):
                    self.add_handler(
                        update_type="initializer",
                        func=func,
                        filters=filters,
                        position=position,
                        inner_object=inner_object,
                        timeout=timeout,
                    )
                else:
                    raise TypeError("Handler must be async")
            elif isinstance(self, pytdbot.filters.Filter):
                func._handler = Handler(
                    func=func,
                    update_type="initializer",
                    filter=self,
                    position=position,
                    inner_object=inner_object,
                    timeout=timeout,
                )
            else:
                func._handler = Handler(
                    func=func,
                    update_type="initializer",
                    filter=filters,
                    position=position,
                    inner_object=inner_object,
                    timeout=timeout,
                )

            return func

        return decorator

    def finalizer(
        self: "pytdbot.Client" = None,
        filters: "pytdbot.filters.Filter" = None,
        position: int = None,
        inner_object: bool = False,
        timeout: float = None,
    ) -> None:
        r"""A decorator to finalize an event object after running all handlers

        Parameters:
            filters (:class:`~pytdbot.filters.Filter`, *optional*):
                An update filter

            position (``int``, *optional*):
                The function position in finalizers list. Default is ``None`` (append)

            inner_object (``bool``, *optional*):
                Wether to pass an inner object of update or not; for example ``UpdateNewMessage.message``. Default is ``False``

            timeout (``float``, *optional*):
                Max execution time for the finalizer before it timeout. Default is ``None``

        Raises:
            :py:class:`TypeError`
        """

        def decorator(func: Callable) -> Callable:
            if hasattr(func, "_handler"):
                return func
            elif isinstance(self, pytdbot.Client):
                if iscoroutinefunction(func):
                    self.add_handler(
                        update_type="finalizer",
                        func=func,
                        filters=filters,
                        position=position,
                        inner_object=inner_object,
                        timeout=timeout,
                    )
                else:
                    raise TypeError("Handler must be async")
            elif isinstance(self, pytdbot.filters.Filter):
                func._handler = Handler(
                    func=func,
                    update_type="finalizer",
                    filter=self,
                    position=position,
                    inner_object=inner_object,
                    timeout=timeout,
                )
            else:
                func._handler = Handler(
                    func=func,
                    update_type="finalizer",
                    filter=filters,
                    position=position,
                    inner_object=inner_object,
                    timeout=timeout,
                )
            return func

        return decorator

    def on_message(
        self: "pytdbot.Client" = None,
        filters: "pytdbot.filters.Filter" = None,
        position: int = None,
        timeout: float = None,
    ) -> None:
        r"""A decorator to handle ``updateNewMessage`` but with ``Message`` object.

        Parameters:
            filters (:class:`~pytdbot.filters.Filter`, *optional*):
                An update filter

            position (``int``, *optional*):
                The function position in handlers list. Default is ``None`` (append)

            timeout (``float``, *optional*):
                Max execution time for the handler before it timeout. Default is ``None``

        Raises:
            :py:class:`TypeError`
        """

        def decorator(func: Callable) -> Callable:
            if hasattr(func, "_handler"):
                return func
            elif isinstance(self, pytdbot.Client):
                if iscoroutinefunction(func):
                    self.add_handler(
                        update_type="updateNewMessage",
                        func=func,
                        filters=filters,
                        position=position,
                        inner_object=True,
                        timeout=timeout,
                    )
                else:
                    raise TypeError("Handler must be async")
            elif isinstance(self, pytdbot.filters.Filter):
                func._handler = Handler(
                    func=func,
                    update_type="updateNewMessage",
                    filter=self,
                    position=position,
                    inner_object=True,
                    timeout=timeout,
                )
            else:
                func._handler = Handler(
                    func=func,
                    update_type="updateNewMessage",
                    filter=filters,
                    position=position,
                    inner_object=True,
                    timeout=timeout,
                )

            return func

        return decorator

    def on_updateScheduledEvent(
        self: "pytdbot.Client" = None,
        filters: "pytdbot.filters.Filter" = None,
        position: int = None,
        timeout: float = None,
    ) -> None:
        r"""A scheduled event has been triggered

        Parameters:
            filters (:class:`~pytdbot.filters.Filter`, *optional*):
                An update filter

            position (``int``, *optional*):
                The function position in handlers list. Default is ``None`` (append)

            timeout (``float``, *optional*):
                Max execution time for the handler before it timeout. Default is ``None``

        Raises:
            :py:class:`TypeError`
        """

        def decorator(func: Callable) -> Callable:
            if hasattr(func, "_handler"):
                return func
            elif isinstance(self, pytdbot.Client):
                if iscoroutinefunction(func):
                    self.add_handler(
                        update_type="updateScheduledEvent",
                        func=func,
                        filters=filters,
                        position=position,
                        inner_object=False,
                        timeout=timeout,
                    )
                else:
                    raise TypeError("Handler must be async")
            elif isinstance(self, pytdbot.filters.Filter):
                func._handler = Handler(
                    func=func,
                    update_type="updateScheduledEvent",
                    filter=self,
                    position=position,
                    inner_object=False,
                    timeout=timeout,
                )
            else:
                func._handler = Handler(
                    func=func,
                    update_type="updateScheduledEvent",
                    filter=filters,
                    position=position,
                    inner_object=False,
                    timeout=timeout,
                )

            return func

        return decorator



from html import escape as _html_escape


def escape_html(text: str, quote: bool = True) -> str:
    r"""Escape HTML characters in the given text

    Parameters:
        text (``str``):
            The text to escape

        quote (``bool``, *optional*):
            Whether to escape double quotes. Default is ``True``

    Returns:
        :py:class:`str`: The escaped text
    """

    return _html_escape(text, quote=quote)


special_chars_v1 = r"_\*`\["
special_chars_v2 = r"\_*[]()~`>#+-=|{}.!"


def escape_markdown(text: str, version: int = 2) -> str:
    r"""Escape Markdown characters in the given text

    Parameters:
        text (``str``):
            The text to escape

        version (``int``, *optional*):
            The Markdown version to escape. Default is ``2``

    Returns:
        :py:class:`str`: The escaped text

    Raises:
        :py:class:`ValueError`: If the given markdown version is not supported
    """

    if version == 1:
        chars = special_chars_v1
    elif version == 2:
        chars = special_chars_v2
    else:
        raise ValueError("Invalid version. Must be 1 or 2.")

    return "".join("\\" + c if c in chars else c for c in text)



__all__ = [
    "RETRY_AFTER_PREFEX",
    "get_running_loop",
    "escape_markdown",
    "escape_html",
    "JSON_ENCODER",
    "json_dumps",
    "json_loads",
    "CallbackData",
    "empty_callback_data",
    "load_callback_data",
    "callback_data",
    "obj_to_json",
    "obj_to_dict",
    "dict_to_obj",
    "create_webapp_secret_key",
    "parse_webapp_data",
    "to_camel_case",
    "create_extra_id",
    "get_bot_id_from_token",
    "get_retry_after_time",
    "bold",
    "italic",
    "underline",
    "strikethrough",
    "spoiler",
    "hyperlink",
    "mention",
    "code",
    "pre",
    "pre_code",
    "quote",
    "rtl",
    "ltr",
]

from .asyncio_utils import get_running_loop
from .escape import escape_markdown, escape_html
from .json_utils import (
    JSON_ENCODER,
    json_dumps,
    json_loads,
    CallbackData,
    empty_callback_data,
    load_callback_data,
    callback_data,
)
from .obj_encoder import obj_to_json, obj_to_dict, dict_to_obj
from .webapps import create_webapp_secret_key, parse_webapp_data
from .strings import (
    RETRY_AFTER_PREFEX,
    to_camel_case,
    create_extra_id,
    get_bot_id_from_token,
    get_retry_after_time,
)
from .text_format import (
    bold,
    italic,
    underline,
    strikethrough,
    spoiler,
    hyperlink,
    mention,
    code,
    pre,
    pre_code,
    quote,
    rtl,
    ltr,
)



try:
    import orjson as json
except ImportError:
    try:
        import ujson as json
    except ImportError:
        import json

from typing import Union

JSON_ENCODER = json.__name__

if JSON_ENCODER == "orjson":

    def json_dumps(
        obj, encode: bool = False, null_terminated: bool = False
    ) -> Union[str, bytes]:
        d = json.dumps(obj)
        if (
            null_terminated
        ):  # Null-terminated string is needed for orjson with c_char_p in tdjson
            d += b"\0"

        return d if encode else d.decode("utf-8")
else:

    def json_dumps(
        obj, encode: bool = False, null_terminated: bool = False
    ) -> Union[str, bytes]:
        d = json.dumps(obj, separators=(",", ":"))
        return d if not encode else d.encode("utf-8")


def json_loads(obj):
    return json.loads(obj)


class CallbackData:
    def __init__(self, action, data=None):
        self.action = action
        self.data = data


empty_callback_data = CallbackData("")


def load_callback_data(data: bytes) -> CallbackData:
    r"""loads already created callback data by :func:`~pytdbot.utils.callback_data`. Returns empty CallbackData on error"""

    if not (data[0] == 0x5B and data[-1] == 0x5D):
        return empty_callback_data

    try:
        d = json_loads(data)
        if not isinstance(d, list) or len(d) != 2:
            return empty_callback_data
    except Exception:
        return empty_callback_data
    else:
        return CallbackData(*d)


def callback_data(action, data=None) -> bytes:
    r"""Create callback data for inline buttons

    Parameters:
        action (``Any``):
            Action name, can be any type, but must be JSON serializable

        data (``Any``, *optional*):
            Callback data, can be any type, but must be JSON serializable

    Returns:
        ``bytes``: Callback data encoded as bytes
    """

    d = json_dumps((action, data), encode=True)
    if len(d) > 64:
        raise ValueError(
            f"Callback data must be less than 64 bytes (got {len(d)} bytes)"
        )

    return d
